
#include <string.h>
#include <io.h>
#include <stdio.h>
#include <fcntl.h>
#include <dir.h>
#include <dos.h>
#include <stdarg.h>

#define byte unsigned char
#define FILENAMELENGTH 13
#define MAXFILES 100
#define MAXFIELDS 512

#define YEAR 0    /* --|                    */
#define MON  1    /*   | for YMD dates      */
#define DAY  2    /* --|                    */
#define byte unsigned char
#define word unsigned int
#define dword unsigned long
#define YES 1
#define NO  0

typedef struct {         /* offset  contents                            */
      byte  version;     /*  0   bitmap field                           */
      byte  date[3];     /*  1   hex YMD                                */
      dword reccount;    /*  4   record count                           */
      word  headersize;  /*  8   size of prologue + field descriptors   */
      word  recsize;     /* 10   size of a record in the file           */
      byte  filler1[2];  /* 12   reserved                               */
      byte  trans_flag;  /* 14   whether transaction is pending         */
      byte  filler2[13]; /* 15   reserved                               */
      byte  mdx_flag;    /* 28   1  "production MDX" (IV only)          */
      byte  filler3[3];  /* 31   reserved                               */
} DBF_PROLOGUE;

typedef struct {        /* offset  contents                             */
      char  name[11];   /*  0   null-terminated field name              */
      byte  type;       /* 11   Logical, Numeric, Memo, Date, Float     */
      byte  filler1[4]; /* 12   reserved                                */
      byte  length;     /* 16   length including decimal point          */
      byte  decimals;   /* 17   size of a record in the file            */
      byte  filler2[13];/* 18   reserved                                */
      byte  mdx_flag;   /* 31   1 = "production MDX" (IV only)          */
} FIELD_DESCRIPTOR;

int pauseflag=NO;
char filetable[MAXFILES][FILENAMELENGTH];
FIELD_DESCRIPTOR fielddescheap[MAXFIELDS];
static char copywrong[]="LISTSTRU Copywrong R. Russell Freeland version 1.0";
static char whuzzup[]="What are you doing in here, anyway?";

void waitkey(void) {
      _AH = 0;                      /* service 0 of BIOS interrupt 16h, */
      geninterrupt(0x16);           /* wait until key struck            */
}

/************************************************************************/
/*  pauseprintf(): keep a line count and pause if flagged;              */
/*                 pass params to printf() in any case                  */
/************************************************************************/

#define PAGESIZE 23

void pauseprintf(char *format,...) {
    va_list argptr;
    static int line_count=0;
    char text[80];

    va_start(argptr,format);      /* using this to pass variable      */
    vsprintf(text,format,argptr); /* # of parameters to pauseprintf() */
    va_end(argptr);
    printf("\n%s",text);
    if (!pauseflag) return;
    line_count++;
    if (line_count>=PAGESIZE) {
        line_count=0;
        printf("\n--more--");
        waitkey();
    }
}


/************************************************************************/
/*  justpath(): isolate just the path portion of a DOS pathname         */
/************************************************************************/


int justpath(char *paramname) {
    char *mark;

    if ((mark=strrchr(paramname,'\\'))!=NULL) *(mark+1)=0;
     else if ((mark=strchr(paramname,':'))!=NULL) *(mark+1)=0;
      else *paramname=0;
    return(strlen(paramname));
}

/************************************************************************/
/*  getfullpath(): expand a filename to include full drive/path         */
/*              ASSUMES THE PASSED STRING IS LONG ENOUGH!!!             */
/*              (requires DOS 3, not sure what happens otherwise)       */
/************************************************************************/

void getfullpathname(char *s) {
    char *mark;

    _ES=_DS=FP_SEG(s);
    _SI=_DI=FP_OFF(s);
    _AH=0x60;
    geninterrupt(0x21);
    mark=strstr(s,"????????");
    if (mark!=NULL) {
        *mark='*';
        strcpy(mark+1,mark+8);
    }
}

/************************************************************************/
/*  getfiles(): expand "wildcard" into full path name, find all         */
/*              matching files and stuff 'em into filetable[],          */
/*              then return the count of files found                    */
/************************************************************************/

int getfiles(char *wildcard, char *ext) {
     /* "wildcard" better be long enough for a serious pathlength since   */
     /* going to expand it with getfullpath()--no way to check though     */
    int a=0,b,done;
    struct ffblk dta;                   /* Turbo C, for findfirst()       */
    char *mark,*fname;

    if ((!strpbrk(wildcard,"\*\?")) &&
        findfirst(wildcard,&dta,16)==0) {
        if (dta.ff_attrib==16) {
              strcat(wildcard,"\\");
        }
    }
    fname=(strrchr(wildcard,'\\'));     /* look for end of path           */
    if (fname==NULL)                    /* if not found then              */
          fname=strrchr(wildcard,':');  /* look for drive (colon)         */
    if (fname==NULL) fname=wildcard;    /* if no drive/path assign wcard  */
     else fname++;                      /* else move past colon/backslash */
    if (strlen(fname)==0)               /* if only a drive/path name,     */
          strcpy(fname,"*");            /* supply the wildcard symbol "*" */
    mark=strchr(fname,'.');
    if (mark==NULL) {                   /* if no "." in name              */
        strcat(wildcard,".");           /* add ".*"                       */
        strcat(wildcard,ext);
    }
    done=(findfirst(wildcard,&dta,39));  /* find any file but directory   */
    if (done) {
        pauseprintf("%s: No matching files found!",wildcard);
        exit(1);
    }
    while (!done) {
        strcpy(filetable[a],dta.ff_name);  /* collect the files so we     */
        a++;                               /* can sort them               */
        if (a==MAXFILES) {
              pauseprintf("%d (maximum) files exceeded!",MAXFILES);
              break;
        }
        done=findnext(&dta);
    }
    if (a>1) {
    /* if we found files matching spec, print count and sort */
        qsort(filetable,a,FILENAMELENGTH,strcmp);
        getfullpathname(wildcard);
        pauseprintf("%d files matching %s:",a,wildcard);
    }
    return(a);
}

/************************************************************************/
/*  printfields(): print the field list from the header buffer          */
/************************************************************************/

void printfields(FIELD_DESCRIPTOR *f,int numfields){
    char *ftypename;
    int counter;
    char is_mdx_flag;
    int totbytes=1;   /* one byte for deleted flag */

    pauseprintf("Field  Field Name  Type       Width    Dec  Index");
    for (counter=1;counter<=numfields;counter++) {
        switch (f->type) {
         case 'C':   ftypename="Character"; break;
         case 'D':   ftypename="Date"; break;
         case 'N':   ftypename="Numeric"; break;
         case 'F':   ftypename="Float"; break;
         case 'L':   ftypename="Logical"; break;
         case 'M':   ftypename="Memo"; break;
         default:    ftypename="Undefined!";
        }
        is_mdx_flag='N';
        if (f->mdx_flag) is_mdx_flag='Y';
        pauseprintf("%5d  %-12s%-10s%6d%7d    %c",
            counter,
            f->name,
            ftypename,
            f->length,
            f->decimals,
            is_mdx_flag);

        totbytes+=f->length;
        f++;
    }
    pauseprintf("**Total record length**%12d",totbytes);
}

void syntaxerror(void) {
    printf("\n\7Syntax: LISTSTRU [filespec] [/P]");
    printf("\nDumps DBF header structure(s) to the screen");
    printf("\nSee Data Based Advisor June 1990 for details");
    exit(1);
}

main(int argc, char *argv[]) {
    char pathname[128],*sumarg;
    int counter,numfiles,pathlen,fhandle,fieldheapsize;
    byte version;
    int foxflag=NO;
    DBF_PROLOGUE prologue;

    _fmode=O_BINARY;
    if (argc>3) syntaxerror();
    for (counter=1;counter<3;counter++) {
        sumarg=argv[counter];
        if (*sumarg=='/') {
            ++sumarg;
            if (*sumarg=='p' || *sumarg=='P') {
                  if (counter==1) argv[1]=argv[2];
                  /* setup filename in first argument */
                  pauseflag=YES;
            }
        }
    }
    if (argc < 2) strcpy(pathname,"*.DBF"); /* if no arguments supply one  */
     else strcpy(pathname,argv[1]);         /* setup local (longer) var    */

    numfiles=getfiles(pathname,"DBF");      /* build sorted table of files */
    pathlen=justpath(pathname);             /* save length of path portion */
    for (counter=0;counter<numfiles;counter++) {
        strcpy(pathname+pathlen,filetable[counter]);
        pauseprintf("");
        pauseprintf("");
        pauseprintf("Structure for database  : %s",pathname);
        if ((fhandle = open(pathname,O_RDONLY+O_BINARY)) == -1) {
            pauseprintf("\7Can't open file!");
            continue;
        }
        if (_read(fhandle,&prologue,
                sizeof(DBF_PROLOGUE))!=sizeof(DBF_PROLOGUE)) {
            pauseprintf("\7Can't read header!");
            _close(fhandle);
            continue;
        }
        version=prologue.version&3;
        if (version==2) {
            pauseprintf("\7dBASE II DBF!  Nahhhh....");
            continue;
        }
        if (prologue.version==0xf5) foxflag=YES;
        if (((prologue.version&3)!=3) && !foxflag) {
            pauseprintf("\7Unrecognized DBF version!");
            continue;
        }
        if (foxflag) {
            pauseprintf("** FoxPro Memo Format **");
        }
         else {
            if (prologue.version&8) {
                pauseprintf("** dBASE IV Memo Format **");
            }
            else {
                pauseprintf("** standard \"III Plus\" Format **");
            }
        }
        pauseprintf("Number of data records  : %d",prologue.reccount);
        pauseprintf("Date of last update     : %d/%d/%d ",
            prologue.date[MON],
            prologue.date[DAY],
            prologue.date[YEAR]);
        fieldheapsize=prologue.headersize-sizeof(DBF_PROLOGUE);
        if (_read(fhandle,fielddescheap,fieldheapsize)!=fieldheapsize) {
            pauseprintf("Can't read field descriptors!");
            _close(fhandle);
            continue;
        }
        printfields(fielddescheap,fieldheapsize/sizeof(FIELD_DESCRIPTOR));
        if (prologue.version&48)  /* if bits 5 & 6 set */
            pauseprintf("dBASE IV SQL file");
         else pauseprintf("Not a dBASE IV SQL file");
        if (prologue.mdx_flag!=0) pauseprintf("Production MDX is attached");
         else pauseprintf("No Production MDX is attached");
        if (prologue.trans_flag!=0) pauseprintf("Transaction is pending");
         else pauseprintf("No Transaction is pending");
        _close(fhandle);
    }
}


